#!/usr/bin/perl
#
# Jonathan Klassen
#
# HMMER_model_maker.pl
# v1.2 - Dec 18, 2012
#
# Pipeline for making hmmer models from a list of input fasta sequences
# Multithreaded, each type of input fasta sequence gets its own thread

use strict;
use warnings;
use Parallel::ForkManager;
use Getopt::Long;

######################################################################################################################
# Processes input arguments
######################################################################################################################

my $usage = "HMMER_model_maker.pl

DEPENDANCIES: MUSCLE, HMMER3 suite, Perl \"Parallel::ForkManager\" module

USAGE:
-c 	number of cores to use						DEFAULT: 1	e.g., perl HMMER_model_maker_v1.pl -i inlist -c 4
-h	displays this usage statement (also using --help)	
-i 	input list of paths to the fasta files to be analyzed		REQUIRED	e.g., perl HMMER_model_maker_v1.pl -i inlist
-q	run quietly, i.e., no STDOUT (Y or N)				DEFAULT: N	e.g., perl HMMER_model_maker_v1.pl -i inlist -q Y
-r	remove intermediate output files (Y or N)			DEFAULT: Y	e.g., perl HMMER_model_maker_v1.pl -i inlist -r N

OUTPUT FILES:
	MUSCLE_aligned_faa_files/<hmm_name>.muscle.faa		A MUSCLE aligned fasta file for each individual protein type; omitted if 
									-q flag set to \"Y\"
	Individual_HMMs/<hmm_name>.hmm				A HMM file for each individual protein type; omitted if -q flag set to \"Y\"
	Mult_HMMs.hmm						A concatenated file containing the HMMs for each individual protein type
	Mult_HMMs.hmm.h3m, .h3i, .h3f, .h3p			The compressed versions of mult_hmms.hmm generated by HMMPRESS
	MUSCLE_aligned_faas.list				A list of all of the MUSCLE aligned fasta files;; omitted if -q flag set to \"Y\"
	HMM_names.list						A list of all of the HMM names
";

# input arguments

my %options = ();
GetOptions(
	"c=i" => \$options{cores}, 		"h" => \$options{help},
	"help" => \$options{help}, 		"i=s" => \$options{infile},
	"q=s" => \$options{quiet},		"r=s" => \$options{remove_intermediates}
);

# display usage statement if called
 
die $usage if ($options{help});

# required arguments

die "Input file is not specified:\n", $usage if (!$options{infile});

# default arguments

unless ($options{cores}){ $options{cores} = 1 }
unless ($options{quiet}){ $options{quiet} = "N" }
unless ($options{remove_intermediates}){ $options{remove_intermediates} = "Y" }

# mystery input flags not allowed

die "Unrecognized command line arguments: @ARGV\n" if ($ARGV[0]);

# checks correct parameter formatting

die "Unrecognized command line arguments: -q = $options{quiet}\n$usage" unless ($options{quiet} eq "Y" or $options{quiet} eq "N");
die "Unrecognized command line arguments: -r = $options{remove_intermediates}\n$usage" 
	unless ($options{remove_intermediates} eq "Y" or $options{remove_intermediates} eq "N");

# checks if input and output files exist

unless (-e $options{infile}){ die "$options{infile} does not exist\n$usage" }
if ($options{quiet} eq "N"){
	if (-e "Mult_HMMs.hmm.h3m"){ print "WARNING: Mult_HMMs HMM files exist and have been overwritten\n" }
	if (-d "MUSCLE_aligned_faa_files"){ print "WARNING: MUSCLE_aligned_faa_files/ output directory exists and has been overwritten\n" }
	if (-d "Individual_HMMs"){ print "WARNING: Individual_HMMs/ output directory exists and has been overwritten\n" }

}
if (-e "Mult_HMMs.hmm.h3m"){ system "rm Mult_HMMs.hmm.h3*" }
if (-d "MUSCLE_aligned_faa_files"){ system "rm -rf MUSCLE_aligned_faa_files*" }
if (-d "Individual_HMMs"){ system "rm -rf Individual_HMMs" }

# print parameters unless -q flag selected

print "
----------------------------------------------------------------
HMMER_model_maker.pl	Jonathan Klassen 	v1.2: Dec 18, 2012

parameters used:
	input file = $options{infile}
	number of cores = $options{cores}
	quiet = $options{quiet}
	remove intermediate output files = $options{remove_intermediates}
----------------------------------------------------------------
" if ($options{quiet} eq "N");

######################################################################################################################
# Loads input files, derives hmm model names from them
######################################################################################################################

my (@files, @names);

open (INLIST, $options{infile}) or die $usage;
while (<INLIST>){
	chomp;
	next unless ($_ =~ /\w/);
	push @files, $_;
	/\/?((\w|-)+)\./ or die "Cannot match input name";
	push @names, $1;
}

######################################################################################################################
# Makes muscle alignments for each multiple fasta files
# Makes hmms for each alignment
######################################################################################################################

mkdir "MUSCLE_aligned_faa_files" or die "Cannot mkdir MUSCLE_aligned_faa_files";
mkdir "Individual_HMMs" or die "Cannot mkdir Individual_HMMs";

##############################################################################
# Multithreaded running of antiSMASH
##############################################################################

sub muscle_hmm_maker($);

our %summary_output;

my $pm = new Parallel::ForkManager($options{cores});

for my $a (0..$#files){
	$pm->start and next;
	my @return = muscle_hmm_maker($a);
	$pm->finish;
}
$pm->wait_all_children();

##############################################################################
# subroutine for making muscle alignments and hmmer models
##############################################################################

sub muscle_hmm_maker ($){
	my $counter = $_[0]; # index counter for global vars
	if ($options{quiet} eq "N"){ print "Aligning $names[$counter]...\n" }
	system "muscle -in $files[$counter] -out MUSCLE_aligned_faa_files/$names[$counter].muscle.faa -quiet";
	if ($options{quiet} eq "N"){ print "Building hmmer model for $names[$counter]...\n" }
	system "hmmbuild --informat afa Individual_HMMs/$names[$counter].hmm MUSCLE_aligned_faa_files/$names[$counter].muscle.faa > $names[$counter].temp.hmmbuild.out";
	system "rm $names[$counter].temp.hmmbuild.out";
}

##############################################################################
# creating final concatenated output files
##############################################################################

if ($options{quiet} eq "N"){ print "\nBuilding outfiles...\n\n" }
system "cat Individual_HMMs/*.hmm > Mult_HMMs.hmm";
system "hmmpress Mult_HMMs.hmm > temp";
system "rm temp";
system "ls MUSCLE_aligned_faa_files/*.faa > MUSCLE_aligned_faas.list";

open (OUTFILE, ">HMM_names.list") or die "Cannot open HMM_names.list";
for my $a (0..$#names){
	print OUTFILE $names[$a], "\n";
}
close OUTFILE;

if ($options{remove_intermediates} eq "Y"){ # Cleans up output files, if requested
	for my $a (0..$#names){
		system "rm -rf Individual_HMMs/";
		system "rm -rf MUSCLE_aligned_faa_files/";
		system "rm -f MUSCLE_aligned_faas.list";
	}
}

	
