#!/usr/bin/perl
#
# Jonathan Klassen
#
# Mult_hmmscan.pl
# v1.202 - Jan. 17, 2017
#
# Searches a set of genome protein translations using the hmms generated using HMMER_model_maker.pl
# Multithreaded, each genome gets its own thread

use strict;
use warnings;
use Parallel::ForkManager;
use Getopt::Long;

######################################################################################################################
# Processes input arguments
######################################################################################################################

my $usage = "Mult_hmmscan.pl

DEPENDANCIES: HMMER3 suite, Perl \"Parallel::ForkManager\" module

USAGE:
-a	input hmm file						DEFAULT: 31_HMMs.hmm		e.g., perl Mult_hmmscan.pl -i inlist -a other1.list
-b	input list of hmm names					DEFAULT: 31_HMM_names.list	e.g., perl Mult_hmmscan.pl -i inlist -b other2.list
	NOTE: This table can be of the format \"<name>\t<evalue>\", specifying the e-value unique to that HMM; otherwise the default set using -e is used
-c 	number of cores to use					DEFAULT: 1			e.g., perl Mult_hmmscan.pl -i inlist -c 4	
-e 	set e-value for parsing the hmmscan results		DEFAULT: 1e-20			e.g., perl Mult_hmmscan.pl -i inlist -e 1e-5
-h	displays this usage statement (also using --help)
-i 	input list of paths to the genome fasta files to be analyzed	REQUIRED		e.g., perl Mult_hmmscan.pl -i inlist
-q	run quitely, i.e., no STDOUT (Y or N)			DEFAULT: N			e.g., perl Mult_hmmscan.pl -i inlist -q Y
-r 	remove intermediate output files (Y or N)		DEFAULT: Y			e.g., perl Mult_hmmscan.pl -i inlist -r N

OUTPUT FILES:
	Mult_hmmscan_match_table.out				A table listing the hmms that were matched (or not) for each genome; 
									changed using the -o flag
	Mult_hmmscan_best_hits					Directory containing summary files of the best hmm hits for each genome
	Mult_hmmscan_individual_faas/<input_genome_name>/	Directory tree containing the individual .faa files for each best hmm hit in 
									each genome; removed using the -r flag
	Mult_hmmscan_hmmscan_outputs				Directory containing the hmmscan output files for each genome; removed using 
									the -r flag
	Mult_hmmscan_multiple_faas				Directory containing the concatenated best hits for each hmm type
	Mult_hmmscan_faa.list					A list of all the multiple fasta files for each hmm type
";

# input arguments

my %options = ();
GetOptions(
	"a=s" => \$options{input_hmm_file}, 		"b=s" => \$options{input_hmm_name_list},
	"c=i" => \$options{cores}, 			"e=s" => \$options{evalue},			
	"h" => \$options{help},				"help" => \$options{help}, 
	"i=s" => \$options{input_genome_faa_list},	"q=s" => \$options{quiet},
	"r=s" => \$options{remove_intermediates}
);

# display usage statement if called
 
die $usage if ($options{help});

# required arguments

die "Input list of genome faas to query is not specified:\n", $usage if (!$options{input_genome_faa_list});

# default arguments

unless ($options{input_hmm_file}){ $options{input_hmm_file} = "31_HMMs.hmm" }
unless ($options{input_hmm_name_list}){ $options{input_hmm_name_list} = "31_HMM_names.list" }
unless ($options{cores}){ $options{cores} = 1 }
unless ($options{evalue}){ $options{evalue} = "1e-20" }
unless ($options{quiet}){ $options{quiet} = "N" }
unless ($options{remove_intermediates}){ $options{remove_intermediates} = "Y" }

# mystery input flags not allowed

die "Unrecognized command line arguments: @ARGV\n" if ($ARGV[0]);

# checks correct parameter formatting

die "Unrecognized command line arguments: -e = $options{evalue}\n$usage" 
	unless ($options{evalue} =~ /^\d+(\.?\d+)?$/ or $options{evalue} =~ /^\d+e-\d+$/);
die "Unrecognized command line arguments: -q = $options{quiet}\n$usage" unless ($options{quiet} eq "Y" or $options{quiet} eq "N");
die "Unrecognized command line arguments: -r = $options{remove_intermediates}\n$usage" 
	unless ($options{remove_intermediates} eq "Y" or $options{remove_intermediates} eq "N");

# checks if input and output files exist

unless (-e $options{input_hmm_file}){ die "$options{input_hmm_file} does not exist\n$usage" }
unless (-e $options{input_genome_faa_list}){ die "$options{input_genome_faa_list} does not exist\n$usage" }
if ($options{quiet} eq "N"){
	if (-e "Mult_hmmscan_match_table.out"){ print "WARNING: Mult_hmmscan_match_table.out exists and has been overwritten\n" }
	if (-e "Mult_hmmscan_faa.list"){ print "WARNING: Mult_hmmscan_faa.list exists and has been overwritten\n" }
	if (-d "Mult_hmmscan_multiple_faas"){ print "WARNING: Mult_hmmscan_multiple_faas output directory exists and has been overwritten\n" }
	if (-d "Mult_hmmscan_hmmscan_outputs"){ print "WARNING: Mult_hmmscan_hmmscan_outputs output directory exists and has been overwritten\n" }
	if (-d "Mult_hmmscan_best_hits"){ print "WARNING: Mult_hmmscan_best_hits output directory exists and has been overwritten\n" }
	if (-d "Mult_hmmscan_individual_faas"){ print "WARNING: Mult_hmmscan_individual_faas output directory exists and has been overwritten\n" }
}
if (-e "Mult_hmmscan_match_table.out"){ system "rm Mult_hmmscan_match_table.out" }
if (-e "Mult_hmmscan_faa.list"){ system "rm Mult_hmmscan_faa.list" }
if (-d "Mult_hmmscan_multiple_faas"){ system "rm -rf Mult_hmmscan_multiple_faas" }
if (-d "Mult_hmmscan_hmmscan_outputs"){ system "rm -rf Mult_hmmscan_hmmscan_outputs" }
if (-d "Mult_hmmscan_best_hits"){ system "rm -rf Mult_hmmscan_best_hits" }
if (-d "Mult_hmmscan_individual_faas"){ system "rm -rf Mult_hmmscan_individual_faas" }

# print parameters unless -q flag selected

print "
----------------------------------------------------------------
Mult_hmmscan.pl		Jonathan Klassen 	v1.2: Dec 18, 2012

parameters used:
	input list of genome faa files to be searched = $options{input_genome_faa_list}
	input hmm file = $options{input_hmm_file}
	input list of hmm names = $options{input_hmm_name_list}
	number of cores = $options{cores}
	minimum evalue for hmm matches = $options{evalue}
	run quietly = $options{quiet}
	remove intermediate files = $options{remove_intermediates}
----------------------------------------------------------------
" if ($options{quiet} eq "N");

# resets output directories

mkdir "Mult_hmmscan_individual_faas";
mkdir "Mult_hmmscan_hmmscan_outputs";
mkdir "Mult_hmmscan_best_hits";

######################################################################################################################
# Loads input files, derives names from them
######################################################################################################################

our (@files, @names, @hmm_names, %hmm_evalues);

open (INLIST, $options{input_genome_faa_list}) or die "Cannot open $options{input_genome_faa_list}";
while (<INLIST>){
	chomp;
	next unless ($_ =~ /\w/);
	push @files, $_;
	/\/?((\w|-)+)\..*$/;
	push @names, $1;
}

close INLIST;

# loads list of hmm names; checks if they have a specified evalues and if not assigns the value set using -e

open (INHMMLIST, $options{input_hmm_name_list}) or die "Cannot open $options{input_hmm_name_list}";
while (<INHMMLIST>){
	chomp;
	my @line = split "\t";
	if (!$line[0]){ $line[0] = $_ }
	push @hmm_names, $line[0];
	if (!$line[1]){
		$hmm_evalues{$line[0]} = $options{evalue};
	}
	elsif ($line[1] =~ /^\d+(\.?\d+)?$/ or $line[1] =~ /^\d+e-\d+$/){
		$hmm_evalues{$line[0]} = sprintf("%e", $line[1]);
	}
	else {
		print "WARNING: $line[1] in $options{input_hmm_name_list} is not recognized as numeric; setting the hmm e-value for $line[0] to $options{evalue}\n";
	}
}
close INLIST;

######################################################################################################################
# Searches each genome using the inputted hmms
# Parses the best hits out of the corresponding .tblout file
######################################################################################################################

sub hmmscan_matcher_and_parser($);

our %summary_output;

my $pm = new Parallel::ForkManager($options{cores});

for my $a (0..$#files){
	$pm->start and next;
	my @return = hmmscan_matcher_and_parser($a);

	unless (-d "temp_outputs"){ mkdir "temp_outputs" }
	open (TEMPOUT, ">temp_outputs/$a.tempout") or die "Cannot open temp_outputs/$a.tempout";
	print TEMPOUT join "\t", @return;
	close TEMPOUT;
	$pm->finish;
}
$pm->wait_all_children();

my @found_summary;
for my $a (0..$#files){
	open (TEMPIN, "temp_outputs/$a.tempout") or die "Cannot open temp_outputs/$a.tempout";
	my @return;
	while (<TEMPIN>){ 
		@return = split "\t";
		shift @return;
		push @found_summary, [@return];
	}
}
system "rm -rf temp_outputs";

######################################################################################################################
# Creates summary and concatenated fasta files
######################################################################################################################	

# summary outfile detailing which hmms had or did not have matches in each genome

open (OUTMATCH, ">Mult_hmmscan_match_table.out") or die "Cannot open outfile Mult_hmmscan_match_table.out";
for my $row (@found_summary){
	print OUTMATCH "@$row\n";
}
close OUTMATCH;

# output directory for multiple fasta files for each hmm

mkdir "Mult_hmmscan_multiple_faas";

# summary outfile listing the multiple fasta files created for each hmm

open (OUTLIST, ">Mult_hmmscan_faa.list") or die "Cannot open outfile Mult_hmmscan_faa.list";
foreach my $hmm_name (@hmm_names){
	system "cat Mult_hmmscan_individual_faas/*/$hmm_name*.faa > Mult_hmmscan_multiple_faas/mult_$hmm_name.faa";
	print OUTLIST "Mult_hmmscan_multiple_faas/mult_$hmm_name.faa\n";
}

######################################################################################################################
# Clean up directories, if required
######################################################################################################################	

if ($options{remove_intermediates} eq "Y"){
	system "rm -rf Mult_hmmscan_individual_faas";
	system "rm -rf Mult_hmmscan_hmmscan_outputs";
	system "rm -rf Mult_hmmscan_best_hits";
}

######################################################################################################################
# Subroutine finding and tallying best hmmscan matches for each genome
######################################################################################################################

sub hmmscan_matcher_and_parser ($){
	my $counter = $_[0]; # index counter for global vars
	my @return; # return array for sub
	my $found = my $not_found = 0;
	$return[0] = $counter; # genome number is the first value returned from the subroutine

	# sets up temporary output directories for each genome

	die "Input file $files[$counter] is not .faa (fasta) format" unless ($files[$counter] =~ /\.faa$/); 
		# only *.faa files are used as search templates
	mkdir "Mult_hmmscan_individual_faas/$names[$counter]";

	# runs hmmscan for the current genome

	if ($options{quiet} eq "N"){ print "Processing genome: $names[$counter]\n" }
	system "hmmscan --tblout Mult_hmmscan_hmmscan_outputs/$names[$counter]\.hmm.tblout -o Mult_hmmscan_hmmscan_outputs/$names[$counter]\.hmm.longout $options{input_hmm_file} $files[$counter]";

	# parses best hmm matches from the hmmscan tblout file

	open (INFILE, "Mult_hmmscan_hmmscan_outputs/$names[$counter]\.hmm.tblout") 
		or die "Cannot open Mult_hmmscan_hmmscan_outputs/$names[$counter]\.hmm.tblout";

	my %outhash;
	while (<INFILE>){
		next if (/^#/);
		my @line = split " ";
		$line[0] =~ s/\.muscle//;
		next if ($line[7] > $hmm_evalues{$line[0]}); 
		if (!$outhash{$line[0]} or $outhash{$line[0]}{'evalue'} > $line[7]){
			$outhash{$line[0]}{'evalue'} = $line[7];
			$outhash{$line[0]}{'hit'} = $line[2]; 
			$outhash{$line[0]}{'file_name'} = $files[$counter];
		}
	}

	# counts how many of the inputs found or not found for the current genome

	my (@founds, @not_founds);
	foreach my $hmm_name (@hmm_names){
		if ($outhash{$hmm_name}){ 
			$found++;
			push @founds, $hmm_name;
		}
		else { 
			$not_found++;
			push @not_founds, $hmm_name;
		}
	}

	# returns summary of those hmms found or not found in that genome

	my $temp_founds = join (" ", @founds);
	my $temp_not_founds = join (" ", @not_founds);
	$return[1] = "$names[$counter]\t$found\t$not_found\tfound:$temp_founds\tnot_found:$temp_not_founds";
	
	######################################################################################################################
	# Constructs summary output files for each genome (deletes if -q eq "Y")
	# Concatenates individual .faa files for each COG
	######################################################################################################################

	# prints summary outfile listing the best hmm hits for the current genome

	open (OUTFILE, ">Mult_hmmscan_best_hits/$names[$counter]\_hmm_best_hits.out") or die 
		"Cannot open Mult_hmmscan_best_hits/$names[$counter]\_hmm_best_hits.out";
	foreach my $COG (sort keys %outhash){
		print OUTFILE "$names[$counter]\t$COG\t$outhash{$COG}{'evalue'}\t$outhash{$COG}{'hit'}\t$outhash{$COG}{'file_name'}\n";
	}
	close OUTFILE;

	# extracts individual fasta files from the original multiple fasta files (only searched once)

	open (INFILE, $files[$counter]) or die "Cannot open $files[$counter]";
	my $found_hit = 0;
	while (<INFILE>){
		if (/>/){
			chomp;
			$found_hit = 0;
			foreach my $hmm_name (sort keys %outhash){
				if (/\Q$outhash{$hmm_name}{'hit'}\E /){
					my $found_hmm_name = $hmm_name;
					my $seq_name;
					# If the header is in GenBank format, only the accession number is kept
					if ($_ =~ /.+\|.+\|.+\|(\w+)(\..*)?\|/){ $seq_name = $1 } 
					# If header has spaces and other special characters in it (e.g., as from MicroScope, IMG), 
					# only data before the first is kept
					elsif ($_ =~ /^>((\w|-)+)/){ $seq_name = $1 }
					else { $seq_name = $_ }
					$found_hit = 1;
					close OUTFILE;
					open (OUTFILE, ">Mult_hmmscan_individual_faas/$names[$counter]/$found_hmm_name\_$names[$counter]\.faa") 
						or die "Cannot open Mult_hmmscan_individual_faas/$names[$counter]/$found_hmm_name\_$names[$counter]\.faa";
					print OUTFILE ">$seq_name $names[$counter] $found_hmm_name\n";
					last;
				}
			}
		}
		elsif ($found_hit == 1){
			s/\*//g; # removes stop codon indicators, if they exist (incompatable with later steps)
			print OUTFILE $_;
		}
	}
	close OUTFILE;
	return @return;
}
